#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <scsi/sg.h>
#include <stdlib.h>


#define SCSI_OFF sizeof(struct sg_header)

static unsigned char cmd[SCSI_OFF + 18];      /* SCSI command buffer */


/* process a complete scsi cmd. Use the generic scsi interface. */

static int handle_scsi_cmd(int fd,
                           unsigned cmd_len,         /* command length */
                           unsigned in_size,         /* input data size */
                           unsigned char *i_buff,    /* input buffer */
                           unsigned out_size,        /* output data size */
                           unsigned char *o_buff     /* output buffer */
                           )
{
  int status=0;
  struct sg_header *sg_hd;

    /* safety checks */
  if(!cmd_len){
    return -1;            /* need a cmd_len != 0 */
  }
  if (!i_buff){
    return -1;             /* need an input buffer != NULL */
  }
  #ifdef SG_BIG_BUFF
    //max buffer = SG_BIG_BUF (8*4096)
  #else
    //max buffer = 4096
  #endif

  if(SCSI_OFF+cmd_len+in_size>(8*4096)){
    return -1;
  }
  if(SCSI_OFF+out_size>(8*4096)){
    return -1;
  }
  if(!o_buff){
    out_size=0;
  }

  /* generic scsi device header construction */
  sg_hd=(struct sg_header *)i_buff;
  sg_hd->reply_len=SCSI_OFF+out_size;
  sg_hd->twelve_byte=cmd_len==12;
  sg_hd->result=0;
  status=write(fd,i_buff,SCSI_OFF+cmd_len+in_size);
  if(status<0||status!=SCSI_OFF+cmd_len+in_size||sg_hd->result){
    /* some error happened */
    fprintf(stderr,"write(generic) result = 0x%x cmd = 0x%x\n",
                    sg_hd->result,i_buff[SCSI_OFF]);
    perror("");
    return status;
  }

  if(!o_buff){
    o_buff=i_buff;       /* buffer pointer check */
  }
  /* retrieve result */
  memset(o_buff,0,SCSI_OFF+out_size);
  status=read(fd,o_buff,SCSI_OFF+out_size);
  if(status<0||status!=SCSI_OFF+out_size||sg_hd->result){
    /* some error happened */
    fprintf(stderr,"read(generic) result = 0x%x cmd = 0x%x\n",
                sg_hd->result,o_buff[SCSI_OFF]);
    fprintf(stderr,"read(generic) sense "
                "%x %x %x %x %x %x %x %x %x %x %x %x %x %x %x %x\n",
                sg_hd->sense_buffer[0],         sg_hd->sense_buffer[1],
                sg_hd->sense_buffer[2],         sg_hd->sense_buffer[3],
                sg_hd->sense_buffer[4],         sg_hd->sense_buffer[5],
                sg_hd->sense_buffer[6],         sg_hd->sense_buffer[7],
                sg_hd->sense_buffer[8],         sg_hd->sense_buffer[9],
                sg_hd->sense_buffer[10],        sg_hd->sense_buffer[11],
                sg_hd->sense_buffer[12],        sg_hd->sense_buffer[13],
                sg_hd->sense_buffer[14],        sg_hd->sense_buffer[15]);
    if(status<0){
      perror("");
    }
  }
  /* Look if we got what we expected to get */
  if(status==SCSI_OFF+out_size){
    status = 0; /* got them all */
  }
  return status;  /* 0 means no error */
}


int reset_device(int fd)
{
  #define RESET_CMD     0x00
  #define RESET_CMDLEN  6
  static unsigned char buffer[SCSI_OFF];
  unsigned char cmdblk [RESET_CMDLEN]={
                  RESET_CMD,        /* command */
                  0,                 /* lun/reserved */
                  0,                 /* reply length 24 bit byte */
                  0,              /* reply length high byte*/
                  0,              /* reply lenght low byte */
                  0                  /* reserved */
  };

  memset(buffer,0,SCSI_OFF);
  memcpy(cmd+SCSI_OFF,cmdblk,sizeof(cmdblk));
  if(handle_scsi_cmd(fd,sizeof(cmdblk),0,cmd,0,0)){
    fprintf( stderr, "Reset failed\n" );
    exit(2);
  }
  return 1;
}




/* request vendor brand and model */
unsigned char *Inquiry(int fd)
{
  #define INQUIRY_CMD     0x12
  #define INQUIRY_CMDLEN  6
  #define INQUIRY_REPLY_LEN 96

  static unsigned char Inqbuffer[ SCSI_OFF + INQUIRY_REPLY_LEN ];
  unsigned char cmdblk[INQUIRY_CMDLEN]={
                  INQUIRY_CMD,        /* command */
                  0,                  /* lun/reserved */
                  0,                  /* page code */
                  0,                  /* reserved */
                  INQUIRY_REPLY_LEN,  /* allocation length */
                  0                   /* reserved/flag/link */ 
  };

  memset(Inqbuffer,0,SCSI_OFF+INQUIRY_REPLY_LEN);
  memset(cmd,0,SCSI_OFF+18);
  memcpy(cmd+SCSI_OFF,cmdblk,sizeof(cmdblk));
  if(handle_scsi_cmd(fd,sizeof(cmdblk),0,cmd,sizeof(Inqbuffer)-SCSI_OFF,Inqbuffer)){
    fprintf( stderr, "Inquiry failed\n" );
    exit(2);
  }
  return (Inqbuffer+SCSI_OFF);
}

int load_unload(int fd)
{
  #define LOAD_CMD     0x1B
  #define LOAD_CMDLEN  6

  //static unsigned char Inqbuffer[ SCSI_OFF + INQUIRY_REPLY_LEN ];
  unsigned char cmdblk[LOAD_CMDLEN]={
                  LOAD_CMD,           /* command */
                  1,                  /* lun/reserved */
                  0,                  /* page code */
                  0,                  /* reserved */
                  0,                  /* allocation length */
                  0                   /* reserved/flag/link */
  };

  memset(cmd,0,SCSI_OFF+18);
  memcpy(cmd+SCSI_OFF,cmdblk,sizeof(cmdblk));
  if(handle_scsi_cmd(fd,sizeof(cmdblk),0,cmd,0,0)){
    fprintf( stderr, "Load/Unload failed\n" );
    exit(2);
  }
  return 1;
}


int TestForMedium (int fd)
{
  #define TESTUNITREADY_CMD 0
  #define TESTUNITREADY_CMDLEN 6
  #define ADD_SENSECODE 12
  #define ADD_SC_QUALIFIER 13
  #define NO_MEDIA_SC 0x3a
  #define NO_MEDIA_SCQ 0x00
  /* request READY status */
  static unsigned char cmdblk[TESTUNITREADY_CMDLEN]={
                      TESTUNITREADY_CMD, /* command */
                      0,                 /* lun/reserved */
                      0,                 /* reserved */
                      0,                 /* reserved */
                      0,                 /* reserved */
                      0                  /* reserved */
  };

  memcpy(cmd+SCSI_OFF,cmdblk,sizeof(cmdblk));
  if(handle_scsi_cmd(fd,sizeof(cmdblk),0,cmd,0,NULL)){
    fprintf (stderr, "Test unit ready failed\n");
    exit(2);
  }
  return
   *(((struct sg_header*)cmd)->sense_buffer+ADD_SENSECODE)!=NO_MEDIA_SC||
   *(((struct sg_header*)cmd)->sense_buffer+ADD_SC_QUALIFIER)!=NO_MEDIA_SCQ;
}




int write_6(int fd,char *p)
{
  #define WRITE_6_CMD     0x0a
  #define WRITE_6_CMDLEN  6
  #define WRITE_6_BUF_LEN 5822
  static unsigned char buffer[SCSI_OFF+WRITE_6_CMDLEN+WRITE_6_BUF_LEN];
  unsigned char cmdblk [WRITE_6_CMDLEN]={
                  WRITE_6_CMD,  /* command */
                  0,            /* lun/reserved */
                  0,            /* buffer length 24 bit byte */
                  0x16,         /* buffer length high byte*/
                  0xbe,         /* buffer length low byte */
                  0             /* reserved */
  };

  memcpy(buffer+SCSI_OFF,cmdblk,sizeof(cmdblk));
  memcpy(buffer+SCSI_OFF+WRITE_6_CMDLEN,p,WRITE_6_BUF_LEN);
  if(handle_scsi_cmd(fd,sizeof(cmdblk),SCSI_OFF+WRITE_6_CMDLEN+WRITE_6_BUF_LEN,buffer,0,0)){
    fprintf( stderr, "Write failed\n" );
    exit(2);
  }
  return 1;
}


int mode_select(int fd,char *p,int size)
{
  #define MODE_SELECT_CMD     0x15
  #define MODE_SELECT_CMDLEN  6
  //static unsigned char buffer[SCSI_OFF+MODE_SELECT_CMDLEN+MODE_SELECT_BUF_LEN];
  unsigned char *buffer;
  unsigned char cmdblk [MODE_SELECT_CMDLEN]={
                  MODE_SELECT_CMD,  /* command */
                  0,            /* SCSI1 (0) or SCSI2 (1) */
                  0,            /* reserved */
                  0,            /* reserved */
                  0,            /* buffer length */
                  0             /* reserved */
  };
  cmdblk[4]=size;
  buffer=(unsigned char *)malloc(SCSI_OFF+MODE_SELECT_CMDLEN+size);
  memcpy(buffer+SCSI_OFF,cmdblk,sizeof(cmdblk));
  memcpy(buffer+SCSI_OFF+MODE_SELECT_CMDLEN,p,size);
  if(handle_scsi_cmd(fd,sizeof(cmdblk),SCSI_OFF+MODE_SELECT_CMDLEN+size,buffer,0,0)){
    fprintf( stderr, "Write failed\n" );
    exit(2);
  }
  free(buffer);
  return 1;
}

unsigned char *read_6(int fd,unsigned char *buf)
{
  #define READ_6_CMD     0x08
  #define READ_6_CMDLEN  6
  #define READ_6_REPLY_LEN 5822
  static unsigned char buffer[SCSI_OFF+READ_6_REPLY_LEN];
  unsigned char cmdblk [READ_6_CMDLEN]={
                  READ_6_CMD,        /* command */
                  0,                 /* lun/reserved */
                  0,                 /* reply length 24 bit byte */
                  0x16,              /* reply length high byte*/
                  0xbe,              /* reply lenght low byte */
                  0                  /* reserved */
  };

  memset(buffer,0,SCSI_OFF+READ_6_REPLY_LEN);
  memcpy(cmd+SCSI_OFF,cmdblk,sizeof(cmdblk));
  if(handle_scsi_cmd(fd,sizeof(cmdblk),0,cmd,sizeof(buffer)-SCSI_OFF,buffer)){
    fprintf( stderr, "Read failed\n" );
    exit(2);
  }
  memcpy(buf,buffer+SCSI_OFF,READ_6_REPLY_LEN);
  return buf;
}

char *mode_sense(int fd,char *p,int size)
{
  #define MODE_SENSE_CMD     0x1A
  #define MODE_SENSE_CMDLEN  6
  unsigned char *buffer;
  unsigned char cmdblk [MODE_SENSE_CMDLEN]={
                  MODE_SENSE_CMD,  /* command */
                  0,            /* SCSI1 (0) or SCSI2 (1) */
                  0,            /* page code */
                  0,            /* reserved */
                  0,            /* buffer length */
                  0             /* reserved */
  };

  cmdblk[4]=size;
  buffer=(unsigned char *)malloc(SCSI_OFF+MODE_SENSE_CMDLEN+size);
  memset(buffer,0,SCSI_OFF+MODE_SENSE_CMDLEN+size);

  memcpy(buffer+SCSI_OFF,cmdblk,sizeof(cmdblk));
  memcpy(buffer+SCSI_OFF+MODE_SELECT_CMDLEN,p,size);

  memcpy(cmd+SCSI_OFF,cmdblk,sizeof(cmdblk));
  if(handle_scsi_cmd(fd,sizeof(cmdblk),0,cmd,MODE_SENSE_CMDLEN+size,buffer)){
    fprintf(stderr,"Write failed\n");
    exit(2);
  }
  memcpy(p,buffer+SCSI_OFF,size);
  free(buffer);
  return p;
}

